/* ---------------------------------------------------------------------------
  TIMER.C    Low-level routines to handle hardware timer interrupt and to
             program hardware clock.
             coolTimer Version 1.21a/C

  Copyright(c) 1993-96 by B-coolWare.  Written by Bobby Z.
 ---------------------------------------------------------------------------*/
/* History:

            Mar 13, 1993  - v1.00
                          initially written.
            Nov 9,  1993  - v1.05
                          added system clock independance, interface slightly
                          changed, new routines added to simplify operation.
            Nov 30, 1993  - v1.06
                          added ExitProc overriding. You do not have to call
                          ResetTimer before exiting anymore - ExitProc will do
                          it by itself.
            Jun 22, 1994  - v1.07
                          changed timer adjust logic - there was a way I
                          missed...
            Jan 27, 1995  - v1.10
                          added two more threads. Still quite precise though.
            Jun 13, 1995  - v1.11
                          C and Pascal versions routine names unified.
            Feb 19, 1996  - v1.20
                          Added more channels, simplified channel addition.
                          More compatibility with multitaskers added (using
                          only documented DOS functions) though I don't think
                          they will allow for precision.
            May 30, 1996  - v1.21
                          fixed one small bug that rendered the system
                          inactive if a handler was assigned a rate less than
                          or equal to 18 TpS AFTER a handler with faster rate
                          was activated.
                          Thanx goes to eXORcist who discovered the bug.
            Jun 27, 1996  - v1.21a
                          removed one far keyword from C version, that caused
                          compilers to give type mismatch error when using
                          small or tiny models. The routine wasn't prototyped
                          as far anyway...
*/

/*
  Note: I'm using assembler almost everywhere because I am sure that almost
        any compiler available can't produce such tightly optimized code as
        I wrote myself. And the speed is significant for this code, thus I
        don't let them 'optimize' it their way. Some code is not necessary to
        be in assembler, but I like it that way - it's much more compact
        anyway. And believe me, if I say it's more compact, I know what I'm
        saying - I compared my assembler code with code generated by most
        compilers available, including Borland C++ 4.52, Turbo C 2.0, Microsoft
        Optimizing C Compiler 5.1 and 7.0, Microsoft Visual C++ 2.2, Watcom
        C/C++ 10.0b and Symantec C++ 6.1. Didn't check with latest versions
        available, but I don't think they did significant improvements in
        optimization anyway :).

  Note 2: for comments on what and how, see TIMER.PAS, the pascal version of
        the coolTimer (which in fact is parent of this C version). Everything
        is described there.
*/

#include "timer.h"
#include <stdlib.h>

      pointer static Save08      = 0L;
      pointer static OldExitProc = 0L;
      uint static TicksPerClk    = 0;
      uint static CurrentTicks   = 0;
      TChannel static Channels[ChannelCount] =
                        {{0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}};

void pascal setTimerSpeed(uint TpS)
{
 CurrentTicks = 0;
 TicksPerClk = ComputeDelay(TpS);
 setTimerDelay(TicksPerClk);
}

void pascal setTimerDelay(uint TDelay)
{
asm {
        cli
        mov     al,36h
        out     43h,al
        mov     ax,TDelay
        out     40h,al
        xchg    ah,al
        out     40h,al
        sti
     }
}

void interrupt internalHandler( void );

void pascal grabTimer( void )
{
 CurrentTicks = TicksPerClk = 0;
asm {
        mov     ax,3508h
        int     21h
        mov     word ptr Save08,bx
        mov     word ptr Save08[2],es
        push    ds
        push    seg internalHandler
        pop     ds
        mov     dx,offset internalHandler
        mov     ax,2508h
        int     21h
        pop     ds
    }
 atexit((void (*)())resetTimer);
}

void pascal resetTimer( void )
{
 register uint i;
 normTimerSpeed();
 for(i = 0; i < ChannelCount; i++ )
  disableHandler(i);
asm {
        mov     si,offset Save08
        mov     ax,word ptr [si]
        or      ax,word ptr [si+2]
        jz      __Q
        push    ds
        mov     dx,word ptr Save08
        mov     ax,word ptr Save08[2]
        mov     ds,ax
        mov     ax,2508h
        int     21h
        pop     ds
    }
__Q:
}

int pascal setTimerHandler(uint Channel, uint Rate, pointer Handler)
{
 if( Channel >= ChannelCount )
  return(ctInvalidChannelIndex);
 Channels[Channel].cTicks = Channels[Channel].cRate = ComputeDelay(Rate);
 Channels[Channel].cHandler = Handler;
 if( TicksPerClk )
  {
   if( Channels[Channel].cRate && (Channels[Channel].cRate < TicksPerClk) )
    setTimerSpeed(Rate);
  }
 else
  setTimerSpeed(Rate);
 return ctOk;
}

void interrupt internalHandler( void )
{
asm {
/*
        mov     ax,seg Channels
        mov     ds,ax
*/
        sub     si,si
    }
__1:
asm {
        mov     ax,word ptr Channels[si].cRate
        test    ax,ax
        jz      __2
        mov     bx,TicksPerClk
        sub     word ptr Channels[si].cTicks,bx
        ja      __2
        add     word ptr Channels[si].cTicks,ax
        mov     ax,word ptr Channels[si].cHandler
        or      ax,word ptr Channels[si].cHandler[2]
        jz      __2
        push    ds
        push    si
        call    dword ptr Channels[si].cHandler
        pop     si
        pop     ds
    }
__2:
asm {
        add     si,TChannelRecSize
        cmp     si,TChannelRecSize*ChannelCount
        jb      __1
        mov     ax,TicksPerClk
        test    ax,ax
        jz      __21
        add     CurrentTicks,ax
        jnc     __11
    }
 __21:
asm {
        sub     si,si
    }
 __22:
asm {
        mov     ax,word ptr Channels[si].cRate
        test    ax,ax
        jnz     __31
        mov     ax,word ptr Channels[si].cHandler
        or      ax,word ptr Channels[si].cHandler[2]
        jz      __31
        push    ds
        push    si
        call    dword ptr Channels[si].cHandler
        pop     si
        pop     ds
    }
__31:
asm {
        add     si,TChannelRecSize
        cmp     si,TChannelRecSize*ChannelCount
        jb      __22

        pushf
        call    dword ptr [Save08]
        jmp     __IRQHandled
        db      10,13,'c','o','o','l','T','i','m','e','r',' ','v','1','.','2','1','/','C'
        db      ' ',' ','C','o','p','y','r','i','g','h','t','(','c',')',' '
        db      '1','9','9','5',',','9','6',' ','b','y',' ','B','-','c','o','o','l','W','a','r','e','.',10,13
    }
 __11:
asm {
        mov     al,20h
        out     20h,al
    }
__IRQHandled:
}

void pascal changeRate(uint Channel, uint Rate)
{
 register pointer Temp;
 if(Channels[Channel].cHandler != 0L)
  {
    Temp = Channels[Channel].cHandler;
    disableHandler(Channel);
    setTimerHandler(Channel, Rate, Temp);
  }
}

uint pascal ComputeDelay(uint TpS)
{
asm {
        mov     bx,TpS
        cmp     bx,18
        mov     ax,0
        jbe     __Q
        mov     dx,12h
        mov     ax,34DDh
        div     bx
    }
__Q:
    return _AX;
}

void pascal playSound(uint Hz)
{
asm {
        mov     bx,Hz
        mov     ax,34DDh
        mov     dx,0012h
        cmp     dx,bx
        jnc     __2
        div     bx
        push    ax
        in      al,61h
        test    al,3
        jnz     __1
        or      al,3
        out     61h,al
        mov     al,0B6h
        out     43h,al
    }
__1:
asm {
        pop     ax
        out     42h,al
        xchg    al,ah
        out     42h,al
    }
__2:
}

void pascal noSound( void )
{
asm {
        in      al,61h
        and     al,0FCh
        out     61h,al
    }
}

void pascal disableHandler(uint Channel)
{
 Channels[Channel].cTicks = Channels[Channel].cRate = 0;
 asm cli;
 Channels[Channel].cHandler = 0L;
 asm sti;
}

void pascal normTimerSpeed(void)
{
asm {
        sub     ax,ax
        push    ax
    }
        setTimerDelay(_AX);
asm {
        pop     ax
        mov     TicksPerClk,ax
        mov     CurrentTicks,ax
        sub     si,si
        mov     cx,ChannelCount
    }
__1:
asm {
        mov     word ptr Channels[si].cTicks,ax
        mov     word ptr Channels[si].cRate,ax
        add     si,TChannelRecSize
        loop    __1
    }
}

